// pastebin.run
// Copyright (C) 2020-2021 Konrad Borowski
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Affero General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Affero General Public License for more details.
//
// You should have received a copy of the GNU Affero General Public License
// along with this program.  If not, see <https://www.gnu.org/licenses/>.

#[macro_use]
extern crate diesel;
#[macro_use]
extern crate rocket;

mod models;
mod routes;
mod schema;

use crate::routes::{config, display_paste, index, insert_paste, metrics, raw_paste};
use chrono::{Duration, Utc};
use diesel::prelude::*;
use models::paste;
use rocket::fairing::AdHoc;
use rocket::http::Header;
use rocket::shield::{Policy, Referrer, Shield};
use rocket::{Build, Ignite, Rocket};
use rocket_dyn_templates::tera::{self, Value};
use rocket_dyn_templates::Template;
use rocket_sync_db_pools::database;
use std::collections::HashMap;
use std::net::SocketAddr;
use tokio::io::{AsyncReadExt, AsyncWriteExt};
use tokio::net::{TcpListener, TcpStream};
use tokio::runtime::{Handle, Runtime};
use tokio::signal::unix::{signal, SignalKind};
use tokio::sync::oneshot::Sender;
use tokio::task::JoinHandle;

#[database("main")]
pub struct Db(PgConnection);

fn js_path(_: &HashMap<String, Value>) -> Result<Value, tera::Error> {
    #[cfg(not(debug_assertions))]
    let path = concat!("/", env!("ENTRY_FILE_PATH"));
    #[cfg(debug_assertions)]
    let path = "http://localhost:5173/js/index.ts";
    Ok(path.into())
}

fn css_stylesheet(_: &HashMap<String, Value>) -> Result<Value, tera::Error> {
    #[cfg(not(debug_assertions))]
    let path = concat!("<link rel=stylesheet href='/", env!("CSS_PATH"), "'>");
    #[cfg(debug_assertions)]
    let path = "";
    Ok(path.into())
}

#[derive(Default)]
struct ContentSecurityPolicy;

impl Policy for ContentSecurityPolicy {
    const NAME: &'static str = "Content-Security-Policy";
    fn header(&self) -> Header<'static> {
        const CONTENT_SECURITY_POLICY: &str = if cfg!(debug_assertions) {
            concat!(
                "default-src 'none';",
                "script-src 'self' localhost:5173;",
                "style-src 'unsafe-inline';",
                "img-src data: https:;",
                "connect-src 'self' ws://localhost:5173;",
                "sandbox allow-forms allow-scripts allow-same-origin;",
                "form-action 'self';",
                "frame-ancestors 'none';",
                "base-uri 'none';",
                "worker-src 'none';",
                "manifest-src 'none'",
            )
        } else {
            concat!(
                "default-src 'none';",
                "script-src 'self';",
                "style-src 'self' 'unsafe-inline';",
                "img-src data: https:;",
                "connect-src 'self';",
                "sandbox allow-forms allow-scripts allow-same-origin;",
                "form-action 'self';",
                "frame-ancestors 'none';",
                "base-uri 'none';",
                "worker-src 'none';",
                "manifest-src 'none'",
            )
        };
        Header::new(Self::NAME, CONTENT_SECURITY_POLICY)
    }
}

fn main() -> Result<(), rocket::Error> {
    let webapp = rocket();

    let (s, r) = tokio::sync::oneshot::channel();

    let runtime = Runtime::new().unwrap();
    runtime.spawn(webapp.launch());

    runtime.spawn(cli());
    runtime.spawn(shutdown_monitor(s));

    let c = runtime.handle().clone();
    runtime.spawn(async move { controller(c) });

    let handle = runtime.handle().clone();
    handle.block_on(async move {
        r.await.unwrap();
        runtime.shutdown_background();
    });

    Ok(())
}

// Auto start after 10s if either service dies
async fn controller(handle: Handle) {
    let mut cli_handle: Option<JoinHandle<()>> = None;
    let mut rocket_handle: Option<JoinHandle<Result<Rocket<Ignite>, rocket::Error>>> = None;
    loop {
        if let Some(h) = cli_handle.as_ref() {
            if h.is_finished() {
                cli_handle = Some(handle.spawn(cli()));
            }
        }
        if let Some(h) = rocket_handle.as_ref() {
            if h.is_finished() {
                rocket_handle = Some(handle.spawn(rocket().launch()));
            }
        }

        use std::time::Duration;
        tokio::time::sleep(Duration::from_millis(10000)).await;
    }
}

async fn shutdown_monitor(s: Sender<i32>) {
    let mut sigint = signal(SignalKind::interrupt()).unwrap();
    let mut sigterm = signal(SignalKind::terminate()).unwrap();

    tokio::select! {
        _ = sigint.recv() => {
            s.send(0).unwrap();
        },
        _ = sigterm.recv() => {
            s.send(0).unwrap();
        },
    }
}

async fn cli() {
    println!("starting cli");
    let addr = "0.0.0.0:9000".parse::<SocketAddr>().unwrap();
    let listener = TcpListener::bind(addr).await.unwrap();

    let mut db = PgConnection::establish(&std::env::var("POSTGRES_URL").unwrap())
        .expect("error in establishing connection with postgresql");

    let host = std::env::var("HOST").unwrap();

    loop {
        let (socket, details) = listener.accept().await.unwrap();

        println!("received connection from {:?}", details);
        save_paste(&mut db, &host, socket).await.unwrap();
    }
}

const BUFFER_SIZE: usize = 128;
const PASTE_SIZE: usize = 2 * 1024 * 1024; // 2MiB

async fn save_paste(
    db: &mut PgConnection,
    host: &str,
    mut socket: TcpStream,
) -> Result<(), Box<dyn std::error::Error>> {
    let mut paste = vec![];
    let mut buffer = [0; BUFFER_SIZE];

    'outer: loop {
        if paste.len() >= PASTE_SIZE {
            break;
        }
        let five_hundred_millis_timeout = tokio::time::Duration::from_millis(500);

        let response =
            tokio::time::timeout(five_hundred_millis_timeout, socket.read(&mut buffer)).await;

        match response {
            Ok(Ok(read)) => {
                if read == 0 {
                    break 'outer;
                }

                paste.extend(buffer);
                buffer.fill(0);
            }
            Ok(Err(e)) => {
                eprintln!("error in reading from socket: {e}");
                break;
            }
            Err(_) => break,
        }
    }

    while let Some(&c) = paste.last() {
        if c == 0x00 {
            paste.pop();
        } else {
            break;
        }
    }

    let string = match String::from_utf8(paste) {
        Ok(v) => v,
        Err(e) => {
            eprintln!("error in parsing string: {}", e);
            return Ok(());
        }
    };

    let v = paste::insert(
        db,
        // Terminal paste should stay for 3 days
        Some(Utc::now() + Duration::days(90)),
        &string,
    )
    .unwrap();

    let output = format!("https://{}/raw/{}\n", host, v);
    let output: Vec<u8> = output.into_bytes();

    socket.write(&output).await?;
    socket.flush().await.unwrap();
    Ok(())
}

fn rocket() -> Rocket<Build> {
    println!("starting rocket");
    let mut rocket = rocket::build()
        .attach(Template::custom(|engines| {
            engines.tera.register_function("js_path", js_path);
            engines
                .tera
                .register_function("css_stylesheet", css_stylesheet);
        }))
        .attach(Db::fairing())
        .attach(AdHoc::on_ignite("Migrations", |rocket| async {
            Db::get_one(&rocket)
                .await
                .expect("a database")
                .run(|conn| diesel_migrations::run_pending_migrations(conn))
                .await
                .expect("database to be migrated");
            rocket
        }))
        .attach(
            Shield::default()
                .enable(ContentSecurityPolicy)
                .enable(Referrer::NoReferrer),
        )
        .mount(
            "/",
            routes![
                config,
                index,
                insert_paste,
                display_paste,
                raw_paste,
                metrics,
            ],
        );
    if cfg!(not(debug_assertions)) {
        rocket = rocket.mount("/assets", rocket::fs::FileServer::from("dist/assets"));
    }
    rocket
}
